package org.implab.gradle.mercurial;

import java.io.Serializable;
import java.util.Objects;
import java.util.Optional;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.annotation.Nonnull;

/**
 * Semantic version object.
 * 
 * <p>
 * This is a readonly version record which follows general rules of
 * semantic versioning. The object is serializable and can be used as
 * input property value for build tasks.
 */
public class SemVersion implements Serializable, Comparable<SemVersion> {

    private static final long serialVersionUID = 1L;

    private static Pattern semVerPattern = Pattern.compile(
            "^(0|[1-9]\\d*)\\.(0|[1-9]\\d*)\\.(0|[1-9]\\d*)(?:-((?:0|[1-9]\\d*|\\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\\.(?:0|[1-9]\\d*|\\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\\+([0-9a-zA-Z-]+(?:\\.[0-9a-zA-Z-]+)*))?$");

    private static Pattern semVerLaxPattern = Pattern.compile(
            "^(0|[1-9]\\d*)(?:\\.(0|[1-9]\\d*)(?:\\.(0|[1-9]\\d*))?)?(?:-((?:0|[1-9]\\d*|\\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\\.(?:0|[1-9]\\d*|\\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\\+([0-9a-zA-Z-]+(?:\\.[0-9a-zA-Z-]+)*))?$");

    private final int major;

    private final int minor;

    private final int patch;

    private final String preRelease;

    private final String meta;

    /** Creates the new and empty version */
    public SemVersion() {
        this(0, 0, 0, null, null);
    }

    /** Creates the version object with the specified values */
    public SemVersion(int major, int minor, int patch, String preRelease, String meta) {
        this.major = major;
        this.minor = minor;
        this.patch = patch;
        this.preRelease = preRelease;
        this.meta = meta;
    }

    /** Creates the version object with the specified values */
    public SemVersion(int major, int minor, int patch) {
        this(major, minor, patch, null, null);
    }

    /** Returns the major part of the version */
    public int getMajor() {
        return major;
    }

    /** Returns the minor part of the version */
    public int getMinor() {
        return minor;
    }

    /** Returns the patch part of the version */
    public int getPatch() {
        return patch;
    }

    /** Gets the optional pre-release version part */
    public Optional<String> getPreRelease() {
        return Optional.ofNullable(preRelease);
    }

    public Optional<String> getMeta() {
        return Optional.ofNullable(meta);
    }

    public boolean isRelease() {
        return getPreRelease().isEmpty();
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();

        sb.append(major).append(".").append(minor).append(".").append(patch);
        getPreRelease().ifPresent(s -> sb.append("-").append(s));
        getMeta().ifPresent(m -> sb.append("+").append(m));

        return sb.toString();
    }

    @Override
    public int compareTo(SemVersion other) {
        int res = Integer.compare(major, other.major);
        if (res != 0)
            return res;
        res = Integer.compare(minor, other.minor);
        if (res != 0)
            return res;
        res = Integer.compare(patch, other.patch);

        return getPreRelease()
                // if we have suffix, check the other version suffix,
                // if the other has no suffix then it's greater.
                // 1.0.0-rc < 1.0.0
                .map(s1 -> other.getPreRelease().map(s2 -> s1.compareTo(s2)).orElse(-1))
                // if we don't have the suffix
                .orElse(other.getPreRelease().isPresent() ? 1 : 0);

    }

    /** Removes the tag (pre-release and meta information) */
    public SemVersion withNoSuffix() {
        return new SemVersion(major, minor, patch, null, null);
    }

    /** Sets the specified suffix leaving other parts unchanged */
    public SemVersion withPreRelease(String preRelease) {
        return new SemVersion(major, minor, patch, preRelease, meta);
    }

    /** Removes pre-release information leaving other parts unchanged */
    public SemVersion withNoPreRelease() {
        return new SemVersion(major, minor, patch, null, meta);
    }

    /** Sets the specified build metadata */
    public SemVersion withMeta(String meta) {
        return new SemVersion(major, minor, patch, preRelease, meta);
    }

    /** Removes the build metadata from the version */
    public SemVersion withNoMeta() {
        return new SemVersion(major, minor, patch, preRelease, null);
    }

    /** Sets the specified patch version */
    public SemVersion withPatch(int patch) {
        return new SemVersion(major, minor, patch, preRelease, meta);
    }

    /** Sets the specified minor version */
    public SemVersion withMinor(int minor) {
        return new SemVersion(major, minor, patch, preRelease, meta);
    }

    /** Sets the specified major version */
    public SemVersion withMajor(int major) {
        return new SemVersion(major, minor, patch, preRelease, meta);
    }

    /** Adds the specified number to the patch version */
    public SemVersion addPatch(int number) {
        return new SemVersion(major, minor, patch + number, preRelease, meta);
    }

    /** Adds the specified number to the minor version */
    public SemVersion addMinor(int number) {
        return new SemVersion(major, minor + number, patch, preRelease, meta);
    }

    /** Adds the specified number to the major version */
    public SemVersion addMajor(int number) {
        return new SemVersion(major + number, minor, patch, preRelease, meta);
    }

    /** Generated the next minor version by incrementing the minor version,
     * setting the patch version to 0 and removing the pre-release information.
     * The build metadata is left unchanged.
     * 
     * <p>Use this method to publish next minor version.
     */
    public SemVersion nextMinor() {
        return new SemVersion(major, minor + 1, 0, null, meta);
    }

    /** Generated the next minor version by incrementing the major version,
     * setting the minor version and the patch to 0 and removing the
     * pre-release information. The build metadata is left unchanged.
     * 
     * <p>Use this method to publish next major version.
     */
    public SemVersion nextMajor() {
        return new SemVersion(major + 1, 0, 0, null, meta);
    }

    /** Parses the version string. */
    public static Optional<SemVersion> tryParse(@Nonnull String str) {
        Objects.requireNonNull(str);

        Matcher match = semVerPattern.matcher(str);

        return match.matches() ? Optional.of(new SemVersion(Integer.parseInt(match.group(1)),
                Integer.parseInt(match.group(2)), Integer.parseInt(match.group(3)), match.group(4), match.group(5)))
                : Optional.empty();
    }

    /** Parses the version string. When  */
    public static Optional<SemVersion> tryParseLax(@Nonnull String str) {
        Objects.requireNonNull(str);

        Matcher match = semVerLaxPattern.matcher(str);

        return match.matches() ? Optional.of(new SemVersion(Integer.parseInt(match.group(1)),
                Integer.parseInt(Optional.ofNullable(match.group(2)).orElse("0")),
                Integer.parseInt(Optional.ofNullable(match.group(3)).orElse("0")), match.group(4), match.group(5)))
                : Optional.empty();
    }

    public static SemVersion toSemVersion(Object value, boolean strict) {
        if (value == null) {
            return new SemVersion();
        } else if (value instanceof SemVersion) {
            return (SemVersion) value;
        } else {
            Optional<SemVersion> semver = strict ? tryParse(value.toString()) : tryParseLax(value.toString());
            return semver.orElseThrow(() -> new IllegalArgumentException(
                    String.format("Failed to parse semver '%s', strict=%s", value, strict)));
        }
    }

    public static SemVersion toSemVersion(Object value) {
        return toSemVersion(value, false);
    }
}
